#!/usr/bin/perl -w

use Encode;
use utf8;

# 2004.09.08 Nobumi Iyanaga
# convert mbox file to utf-8 file
# first  arg: input mbox file;
# second arg: output file;
# third  arg: index file;
# fourth arg: "from" or "to"

#@ARGV = ("/Volumes/myDocument/Users/ni/Desktop/untitled folder/matsumoto_mbox", "/Volumes/myDocument/Users/ni/Desktop/matsumoto_mbox.txt", "/Volumes/myDocument/Users/ni/Desktop/matsumoto_mail_index2.txt", "from");

my $inputfile = shift;
my $outputfile = shift;
my $index_file = shift;

my $from_or_to = shift;

#$inputfile = "/Volumes/myDocument/Users/ni/Desktop/untitled folder/testmail2.txt";
#open (OUT, ">/Volumes/myDocument/Users/ni/Desktop/mailres5.txt");

open (IN, $inputfile) || die "Couldn't open input file $inputfile: $!";

open (OUT, '>:utf8', $outputfile) || die "Couldn't open output file $outputfile: $!";

# binmode(STDOUT, ":utf8");

open (INDICE, '>:utf8', $index_file) || die "Couldn't open index file $index_file: $!";

my ($mail, $mail_begins, $mailtext, $time_from_to, $now);

while (<IN>) {
	if (/^From .+@?.+ \d{2}:\d{2}:\d{2} \d{4}$/) {
		$mail_begins = 1;
		if ($mail) {
			($mailtext, $time_from_to) = &process_mail ($mail, $from_or_to);
			$now = localtime;
			print INDICE "Processing: $time_from_to at $now\n";
			print OUT $mailtext, "\n";
		}
		$mail = "";
	}
	if ($mail_begins) {
		$mail .= $_;
	}
}

if ($mail) {
	($mailtext, $time_from_to) = &process_mail ($mail, $from_or_to);
	$now = localtime;
	print INDICE "Processing: $time_from_to at $now\n";
	print OUT $mailtext, "\n";
}


sub process_mail {
	my $mail = shift;
	my $from_or_to = shift;
	
	my $res = "";
	my $time_from_to;
	my @parts;
	
	my ($header, $body) = $mail =~ /^(.+?\n\n)(.+)$/s;
	
	if (! $header) {
		$mail =~ s/(\x1B\$B[^\x1B]+\x1B\(?B?)/&jis2utf8 ($1)/egs;
		if ($mail =~ /\n(.+=\?)/) {
			$mail =~ s/\n([^\n]+=\?[^\n]*)/"\n". &decode_mimeHeader ($1)/egs;
		}
		if ($mail =~ /^From /) {
			if (! $from_or_to) {
				$from_or_to = "from";
			}
			$time_from_to = &time_from_to ($mail, $from_or_to);
		}
		return ($mail, $time_from_to);
	}
	
#	print "header: $header\nbody: $body\n";
#	return;
	
	$header =~ s/(\x1B\$B[^\x1B]+\x1B\(?B?)/&jis2utf8 ($1)/egs;
	
	if ($header =~ /\n(.+=\?)/) {
		$header =~ s/\n([^\n]+=\?[^\n]*)/"\n". &decode_mimeHeader ($1)/egs;
	}
	
	if ($header =~ /^From /) {
		if (! $from_or_to) {
			$from_or_to = "from";
		}
		$time_from_to = &time_from_to ($header, $from_or_to);
	}
# 	my $now = localtime;
#  	print "Processing: $time_from_to at $now\n";
#	return;
	
	if ($header =~ /boundary="?([^"\n]+)"?/) { # it's a multipart mail...
		my (@boundaries) = $mail =~ /boundary="?([^"\n]+)"?/ig;
		@parts = &parseParts ($body, @boundaries);
		for my $part (@parts) {
# 			print "part$j\n$part";
			my ($part_header, $part_body) = $part =~ /^(.+?\n\n)(.*)$/s;
			if (! $part_header) { # || ! $part_body) {
				$res .= $part;
#				print $part
			}
			else {
				if ($part_header =~ /\n(.+=\?)/) {
					$part_header =~ s/\n([^\n]+=\?[^\n]*)/"\n". &decode_mimeHeader ($1)/egs;
				}
#				print "header\n$header\n\nbody\n$body\n";
				my ($contenttype, $charset, $tr_encoding) = &parseHeader ($part_header);
#				print "contenttype: $contenttype\ncharset: $charset\ntr_encoding: $tr_encoding\n\n";
				$res .= $part_header . &convert_part ($part_body, $contenttype, $tr_encoding, $charset);
			}
		}
		
		$res = $header . $res;
#		print $res;
		return ($res, $time_from_to);
#		return;
	}
		
		
	my ($contenttype, $charset, $tr_encoding) = &parseHeader ($header);
#	print "contenttype: $contenttype\ncharset: $charset\ntr_encoding: $tr_encoding\n\n";
#	return;
#	print "header\n$header\nbody\n$body\n";
#	return;
	$res = $header . &convert_part ($body, $contenttype, $tr_encoding, $charset);
	
#	print $res;

	return ($res, $time_from_to);
	
}

sub convert_part {
	my $str = shift;
	my $contenttype = shift;
	my $tr_encoding = shift;
	my $charset = shift;
	
	my $res = "";
	
	if (! $str) {
		return "";
	}
	
	if ($contenttype && $contenttype =~ /appl|audio|image|attachment/i) {
		return ("***removed data***\n");
	}
	if ((! $contenttype && ! $charset && ! $tr_encoding) || ($contenttype && ! $charset && ! $tr_encoding) || ($tr_encoding && $tr_encoding eq "binary")) {
		if ($str =~ /(\x1B\$B[^\x1B]+\x1B\(?B?)/) {
			$str =~ s/(\x1B\$B[^\x1B]+\x1B\(?B?)/&jis2utf8 ($1)/egs;
		}
		return ($str);
	}
	
	elsif ($tr_encoding) {
		$res = &convert_body ($str, $tr_encoding, $charset);
	}
	elsif ($charset) {
		my $try = "\$res = decode (\$charset, \$str)";
		eval $try;
		$res = $str if $@;
	}
	
	if ($res =~ /\&[0-z]+?;/) {
		$res =~ s/^(.+)$/&htmlEntity ($1)/egs;
	}
	
	return ($res);
}

sub parseHeader {
	my $header = shift;
	
	my ($contenttype, $charset, $tr_encoding, $dummy);
	my ($temp);
	
	if ($header =~ /\n(Content\-Type:[^\n]+)/is) {
		($contenttype) = $header =~ /\n(Content\-Type:[^\n]+)/is;
		($dummy, $temp) = $contenttype =~ /(Content\-Type:\s*)([^\s;]+)/i;
		($dummy, $charset) = $contenttype =~ /(charset="?)([^";]+)"?/i;
		$contenttype = $temp;
	}
	if (! $charset) {
		($dummy, $charset) = $header =~ /\n\s*(charset="?)([^"\n;]+"?)/is;
	}
	if ($header =~ /\n(Content\-Transfer\-Encoding:\s*)([^\s;]+)/is) {
		($dummy, $tr_encoding) = $header =~ /\n(Content\-Transfer\-Encoding:\s*)([^\s;]+)/is;
	}
	
	if ($charset) {
		$charset = uc $charset;
		$charset =~ s/[" ]//g;
		if ($charset =~ /3Diso\-8859\-=/i) {
			$charset = "iso-8859-1";
		}

	}
	
	return ($contenttype, $charset, $tr_encoding);
}

sub time_from_to {
	my $header = shift;
	my $to_or_from = shift;
	
	my $res;
	
	my ($month, $day, $time, $year) = $header =~ /^From .+ (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec) ( ?\d{1,2}) (\d{2}:\d{2}:\d{2}) (\d{4})/;
	
	my $date = &the_date ($month, $day, $time, $year);	
	
	my ($from) = $header =~ /\nFrom: ([^\n]+)/s;
	if (! $from) {
		$from = "xxx";
	}
	else {
		$from =~ s/"|,.*//g;
		$from =~ s/^(.+)( <.+)$/$1/;
		$from =~ s/,? /_/g;
	}
	
	my ($to) = $header =~ /\nTo: ([^\n]+)/s;
	if (! $to) {
		$to = "xxx";
	}
	else {
		$to =~ s/"|,.*//g;
		$to =~ s/^(.+)( <.+)$/$1/;
		$to =~ s/,? /_/g;
	}
	
	if ($to_or_from eq "to") {
		$res = $date . "_" . $to;
	}
	else {
		$res = $date . "_" . $from;
	}
	
	$res =~ s/ |,|, |:|\//_/g;
	
	return ($res);
}

sub parseParts {
	my ($body, @boundaries) = @_;
	
	my (@parts);
	my $i = 0;
	for my $boundary (@boundaries) {
		$boundaries[$i] = quotemeta ("--" . $boundary) . '(\-\-)?';
		$i++;
	}
	
	while ($body) {
		if ($body !~ /$boundaries[0]/) {
			push (@parts, $body);
			$body = "";
		}
	
		my @dummies;
		my @lengthes;
		my $dummy = "dummy";
		$i = 0;
		for $boundary (@boundaries) {
			$dummy .= $i;
			($$dummy) = $body =~ /^(.+?)(\n$boundary)/s;
			if ($$dummy) {
				push (@dummies, $$dummy);
			}
		}
		if (@dummies) { # get the shortest part...
			for my $thispart (@dummies) {
				push (@lengthes, length ($thispart));
			}
			my $nth = &minimum (@lengthes);
			$part = $dummies[$nth];
			push (@parts, $part);
			my $quotedpart = quotemeta ($part);
			$body =~ s/$quotedpart//s;
			
			@lengthes = ();
			@dummies = ();
		}
		
		else {
			if ($body =~ /^(\n*)$boundaries[0](\n*)$/s) {
				$part = $body;
				push (@parts, $part);
				$body = "";
			}					
			my $temp = $body	;
			$temp =~ s/^(\n*)$boundaries[0](\n*)//s;
			if ($temp !~ /$boundaries[0]/s) {
				$part = $body;
				push (@parts, $part);
				$body = "";
			}
		}
	
		if (! $part) {
			push (@parts, $body);
			$body = "";
		}
	}
	
	return (@parts);
} 


sub minimum { # returns the nth of the minimum of a list of numbers
	my @num = @_;

	my $res_num;

	my $min = shift @num;
	my $min_num = 0;

	for my $thenum (@num) {
		$min_num++;
		if ($thenum < $min) {
			$min = $thenum;
			$res_num = $min_num;
		}
	}
	$res_num = 0 if ! $res_num;

	return ($res_num);
}

sub splitPlus { # not used any longer...
	$delim = shift;
	$str = shift;
	
	my @res_array;

	@temp = split ($delim, $str);

	$res_array[0] = shift (@temp);

	for $i (@temp) {
		$res_array[++$#res_array] = $delim . $i;
	}

	return (@res_array);
}

sub convert_body {
	my $str = shift;
	my $enc = shift;
	my $charset = shift;
	
	use MIME::Base64;
	use MIME::QuotedPrint;
	
	$enc = lc ($enc);
	
	if ($enc eq "binary") {
		return ($str);
	}
	elsif ($enc eq "8bit" && ! $charset) {
		$str = decode ("iso-8859-1", $str);
		return ($str);
	}
	elsif ($enc eq "quoted-printable") {
		$str = decode_qp ($str);
	}
	elsif ($enc eq "base64") {
#		print $str, "\n";
		if ($str =~ /^\n+$/) {
			return ($str);
		}
# 		$str =~ s/^(.*)\n\n(.+)$/$2/s;
# 		my $header = $1;
# 		$header =~ s/\n([^\n]+=\?[^\n]*)/"\n". &decode_mimeHeader ($1)/egs if $header;
# 		if ($header && $header =~ /attachment/i) {
# 			return ($header . "\n\n" . $str);
# 		}
		$str = decode_base64($str);
		$str = $header . "\n\n" . $str if $header;
	}
	
	if ($charset) {
		my $try = "\$str = decode (\$charset, \$str)";
		eval $try;
		return ($str);
	}
	
	else {
		if ($str =~ /(\x1B\$B[^\x1B]+\x1B\(?B?)/) {
			$str = &jis2utf8 ($str);
		}
		
		return ($str);
	}
	
}

sub the_date {
	my ($month, $day, $time, $year) = @_;
	
	my %months = ("Jan" => 1, "Feb" => 2, "Mar" => 3,
	"Apr" => 4, "May" => 5, "Jun" => 6, "Jul" => 7,
	"Aug" => 8, "Sep" => 9, "Oct" => 10, "Nov" => 11,
	"Dec" => 12);
	
	if (exists $months{$month}) {
		$month = $months{$month};
	}
	
	return ("$year-$month-$day-$time");
}

sub jis2utf8 {
	$str = shift;
	
	$str = decode ("iso-2022-jp", $str);
	
	return ($str);
}

sub decode_mimeHeader {
	$str = shift;
	
	use MIME::Base64;
	use MIME::QuotedPrint;

	my ($first_part, $char_set, $the_encoding, $thestr, $my_quote, $res);
	
	while ($str) {
		if ($str =~ /=\?/) {
			 ($first_part, $char_set, $the_encoding, $thestr, $my_quote) = $str =~ /^(.*?"?)=\?([^?]+)\?([BbQq])\?([^?]+)\?=("?)/;
			return ($str) if ! $the_encoding;
			if ($the_encoding eq "B" || $the_encoding eq "b") {
				$thestr = decode_base64 ($thestr);
			}
			if ($the_encoding eq "Q" || $the_encoding eq "q") {
				$thestr = decode_qp ($thestr);
			}
			
			my $try = "\$thestr = decode (\$char_set, \$thestr)";
			eval $try;
			
			if ($first_part) {
				$res .= $first_part . $thestr;
			}
			if ($my_quote) {
				$res = $res . '"';
			}
			
			$str =~ s/^(.*?"?)=\?([^?]+)\?([BbQq])\?([^?]+)\?="?//;
		}
		else {
			last;
		}
	}
	
	if ($str) {
		$res .= $str;
	}
#	print $char_set, "\n", $thestr, "\n";
		
	return ($res);
}

sub htmlEntity {
	$str = shift;
	
	$str =~ s/&nbsp;/pack ("U0U", 160)/eg;
	$str =~ s/&iexcl;/pack ("U0U", 161)/eg;
	$str =~ s/&cent;/pack ("U0U", 162)/eg;
	$str =~ s/&pound;/pack ("U0U", 163)/eg;
	$str =~ s/&curren;/pack ("U0U", 164)/eg;
	$str =~ s/&yen;/pack ("U0U", 165)/eg;
	$str =~ s/&brvbar;/pack ("U0U", 166)/eg;
	$str =~ s/&sect;/pack ("U0U", 167)/eg;
	$str =~ s/&uml;/pack ("U0U", 168)/eg;
	$str =~ s/&copy;/pack ("U0U", 169)/eg;
	$str =~ s/&ordf;/pack ("U0U", 170)/eg;
	$str =~ s/&laquo;/pack ("U0U", 171)/eg;
	$str =~ s/&not;/pack ("U0U", 172)/eg;
	$str =~ s/&shy;/pack ("U0U", 173)/eg;
	$str =~ s/&reg;/pack ("U0U", 174)/eg;
	$str =~ s/&macr;/pack ("U0U", 175)/eg;
	$str =~ s/&deg;/pack ("U0U", 176)/eg;
	$str =~ s/&plusmn;/pack ("U0U", 177)/eg;
	$str =~ s/&sup2;/pack ("U0U", 178)/eg;
	$str =~ s/&sup3;/pack ("U0U", 179)/eg;
	$str =~ s/&acute;/pack ("U0U", 180)/eg;
	$str =~ s/&micro;/pack ("U0U", 181)/eg;
	$str =~ s/&para;/pack ("U0U", 182)/eg;
	$str =~ s/&middot;/pack ("U0U", 183)/eg;
	$str =~ s/&cedil;/pack ("U0U", 184)/eg;
	$str =~ s/&sup1;/pack ("U0U", 185)/eg;
	$str =~ s/&ordm;/pack ("U0U", 186)/eg;
	$str =~ s/&raquo;/pack ("U0U", 187)/eg;
	$str =~ s/&frac14;/pack ("U0U", 188)/eg;
	$str =~ s/&frac12;/pack ("U0U", 189)/eg;
	$str =~ s/&frac34;/pack ("U0U", 190)/eg;
	$str =~ s/&iquest;/pack ("U0U", 191)/eg;
	$str =~ s/&Agrave;/pack ("U0U", 192)/eg;
	$str =~ s/&Aacute;/pack ("U0U", 193)/eg;
	$str =~ s/&Acirc;/pack ("U0U", 194)/eg;
	$str =~ s/&Atilde;/pack ("U0U", 195)/eg;
	$str =~ s/&Auml;/pack ("U0U", 196)/eg;
	$str =~ s/&Aring;/pack ("U0U", 197)/eg;
	$str =~ s/&AElig;/pack ("U0U", 198)/eg;
	$str =~ s/&Ccedil;/pack ("U0U", 199)/eg;
	$str =~ s/&Egrave;/pack ("U0U", 200)/eg;
	$str =~ s/&Eacute;/pack ("U0U", 201)/eg;
	$str =~ s/&Ecirc;/pack ("U0U", 202)/eg;
	$str =~ s/&Euml;/pack ("U0U", 203)/eg;
	$str =~ s/&Igrave;/pack ("U0U", 204)/eg;
	$str =~ s/&Iacute;/pack ("U0U", 205)/eg;
	$str =~ s/&Icirc;/pack ("U0U", 206)/eg;
	$str =~ s/&Iuml;/pack ("U0U", 207)/eg;
	$str =~ s/&ETH;/pack ("U0U", 208)/eg;
	$str =~ s/&Ntilde;/pack ("U0U", 209)/eg;
	$str =~ s/&Ograve;/pack ("U0U", 210)/eg;
	$str =~ s/&Oacute;/pack ("U0U", 211)/eg;
	$str =~ s/&Ocirc;/pack ("U0U", 212)/eg;
	$str =~ s/&Otilde;/pack ("U0U", 213)/eg;
	$str =~ s/&Ouml;/pack ("U0U", 214)/eg;
	$str =~ s/&times;/pack ("U0U", 215)/eg;
	$str =~ s/&Oslash;/pack ("U0U", 216)/eg;
	$str =~ s/&Ugrave;/pack ("U0U", 217)/eg;
	$str =~ s/&Uacute;/pack ("U0U", 218)/eg;
	$str =~ s/&Ucirc;/pack ("U0U", 219)/eg;
	$str =~ s/&Uuml;/pack ("U0U", 220)/eg;
	$str =~ s/&Yacute;/pack ("U0U", 221)/eg;
	$str =~ s/&THORN;/pack ("U0U", 222)/eg;
	$str =~ s/&szlig;/pack ("U0U", 223)/eg;
	$str =~ s/&agrave;/pack ("U0U", 224)/eg;
	$str =~ s/&aacute;/pack ("U0U", 225)/eg;
	$str =~ s/&acirc;/pack ("U0U", 226)/eg;
	$str =~ s/&atilde;/pack ("U0U", 227)/eg;
	$str =~ s/&auml;/pack ("U0U", 228)/eg;
	$str =~ s/&aring;/pack ("U0U", 229)/eg;
	$str =~ s/&aelig;/pack ("U0U", 230)/eg;
	$str =~ s/&ccedil;/pack ("U0U", 231)/eg;
	$str =~ s/&egrave;/pack ("U0U", 232)/eg;
	$str =~ s/&eacute;/pack ("U0U", 233)/eg;
	$str =~ s/&ecirc;/pack ("U0U", 234)/eg;
	$str =~ s/&euml;/pack ("U0U", 235)/eg;
	$str =~ s/&igrave;/pack ("U0U", 236)/eg;
	$str =~ s/&iacute;/pack ("U0U", 237)/eg;
	$str =~ s/&icirc;/pack ("U0U", 238)/eg;
	$str =~ s/&iuml;/pack ("U0U", 239)/eg;
	$str =~ s/&eth;/pack ("U0U", 240)/eg;
	$str =~ s/&ntilde;/pack ("U0U", 241)/eg;
	$str =~ s/&ograve;/pack ("U0U", 242)/eg;
	$str =~ s/&oacute;/pack ("U0U", 243)/eg;
	$str =~ s/&ocirc;/pack ("U0U", 244)/eg;
	$str =~ s/&otilde;/pack ("U0U", 245)/eg;
	$str =~ s/&ouml;/pack ("U0U", 246)/eg;
	$str =~ s/&divide;/pack ("U0U", 247)/eg;
	$str =~ s/&oslash;/pack ("U0U", 248)/eg;
	$str =~ s/&ugrave;/pack ("U0U", 249)/eg;
	$str =~ s/&uacute;/pack ("U0U", 250)/eg;
	$str =~ s/&ucirc;/pack ("U0U", 251)/eg;
	$str =~ s/&uuml;/pack ("U0U", 252)/eg;
	$str =~ s/&yacute;/pack ("U0U", 253)/eg;
	$str =~ s/&thorn;/pack ("U0U", 254)/eg;
	$str =~ s/&yuml;/pack ("U0U", 255)/eg;

	return ($str);
}
